-- name: \\#00ff00\\Quickplay
-- incompatible: gamemode
-- description: A streamlined time attack gamemode with emphasis on player coordination. Collect all the stars in each course, either solo or with a party!\n\nFEATURES\n- Level Select\n- All Acts Combined\n- All Players Together\n- Saves Record Times (Solo Only)\n- Several QoL Changes\n- ROM Hack Support (Star Road, SM74, Sapphire, Rainbow Road)\n\nCREDITS\n- Written and Designed by Dan\n- PeachyPeach and Isaac (Object Toggle / Progression)\n- Sunk (Blue Coin Switch Fix, Star Cutscene Skip)\n- EmilyEmmi (DDD Submarine)\n- Djoslin (Cannon Timer)\n- wereyoshi (Race Resets, TTM Monkey Fix)\n- birdkek (HUD Name Processing)

gGlobalSyncTable.level = 0
gGlobalSyncTable.timer = 0
gGlobalSyncTable.transform = false
gGlobalSyncTable.reload = false
gGlobalSyncTable.levelIndex = 1

players = 0

fullGameIndex = 1
soloRun = true
autoRepeat = 0
autoRepeatMessage = ""

local syncCooldown = 900
local reloadTimer = 0


local function update()

    -- Force Players in Level
    if
        gGlobalSyncTable.level == 0 and
        gNetworkPlayers[0].currLevelNum ~= waitingRoom
    then
        warp_to_level(waitingRoom, 1, 0)
    end

    if gGlobalSyncTable.reload then
        warp_to_level(waitingRoom, 1, 0)
    elseif
        gGlobalSyncTable.level > 0 and
        gNetworkPlayers[0].currLevelNum ~= levels[gGlobalSyncTable.level][1]
    then
        warp_to_level(levels[gGlobalSyncTable.level][1], 1, 0)
    end

    -- Get Player Count
    players = 0
    for i=0, MAX_PLAYERS-1 do
        if gNetworkPlayers[i].connected then
            players = players + 1
        end
    end

    -- Host
    if network_is_server() then

        -- Desync Fix Attempt
        if syncCooldown <= 0 then
            for i=1,(MAX_PLAYERS-1) do
                if gNetworkPlayers[i].connected then
                    network_send_to(i, true,
                    {
                        packet = "sync",
                        level = gGlobalSyncTable.level,
                        timer = gGlobalSyncTable.timer,
                        transform = gGlobalSyncTable.transform,
                        reload = gPlayerSyncTable.reload,
                        levelIndex = gGlobalSyncTable.levelIndex,
                    }
                    )
                end
            end
            syncCooldown = 900
        else
            syncCooldown = syncCooldown - 1
        end

        -- Get Host Input
        local input = gMarioStates[0].controller.buttonDown

        if gGlobalSyncTable.level == 0 then -- INTERMISSION

            -- Level Select UI
            if input == L_JPAD then
                autoRepeat = autoRepeat + 1
                if autoRepeat == 1 or autoRepeat > 7 then
                    if gGlobalSyncTable.levelIndex <= 0 then
                        gGlobalSyncTable.levelIndex = #levels
                        play_sound(SOUND_MENU_CHANGE_SELECT, gMarioStates[0].marioObj.header.gfx.cameraToObject)
                    else
                        gGlobalSyncTable.levelIndex = gGlobalSyncTable.levelIndex - 1
                        play_sound(SOUND_MENU_CHANGE_SELECT, gMarioStates[0].marioObj.header.gfx.cameraToObject)
                    end
                    if autoRepeat > 7 then
                        autoRepeat = 5
                    end
                end
            elseif input == R_JPAD then
                autoRepeat = autoRepeat + 1
                if autoRepeat == 1 or autoRepeat > 7 then
                    if gGlobalSyncTable.levelIndex >= #levels then
                        gGlobalSyncTable.levelIndex = 0
                        play_sound(SOUND_MENU_CHANGE_SELECT, gMarioStates[0].marioObj.header.gfx.cameraToObject)
                    else
                        gGlobalSyncTable.levelIndex = gGlobalSyncTable.levelIndex + 1
                        play_sound(SOUND_MENU_CHANGE_SELECT, gMarioStates[0].marioObj.header.gfx.cameraToObject)
                    end
                    if autoRepeat > 7 then
                        autoRepeat = 5
                    end
                end
            else
                autoRepeat = 0
            end

            -- Press Y to Start Round
            if input == Y_BUTTON then
                start_round()
            end

        else --ACTIVE ROUND

            -- Increment Timer
            gGlobalSyncTable.timer = gGlobalSyncTable.timer + 1

            -- Solo Run Validation Check
            if players > 1 then
                soloRun = false
            end
            
            -- Reload
            if gGlobalSyncTable.reload then
                reloadTimer = reloadTimer - 1
                if reloadTimer <= 0 then
                    gGlobalSyncTable.reload = false
                end
            end

            -- Round Clear Condition
            if all_stars_collected() then
                if gGlobalSyncTable.levelIndex == 0 and fullGameIndex < #fullGameSeq then
                    next_level()
                else
                    end_round()
                    record_time()
                end
            end

            -- Input Handlers
            if input == D_JPAD then
                autoRepeatMessage = "Quit Round..."
                autoRepeat = autoRepeat + 1
                if autoRepeat > 15 then
                    end_round()
                end
            elseif input == U_JPAD and not gGlobalSyncTable.reload then
                autoRepeatMessage = "Reload Course..."
                autoRepeat = autoRepeat + 1
                if autoRepeat > 15 then
                    reloadTimer = 30
                    gGlobalSyncTable.reload = true
                    autoRepeat = 0
                end
            else
                autoRepeat = 0
            end

        end

    end

end

function start_round()
    -- Reset save for all players
    packet_receive({packet="reset"})
    network_send(true, {packet="reset"})
    -- Set Variables
    gGlobalSyncTable.timer = 0
    gGlobalSyncTable.transform = false
    soloRun = true
    -- Set Level
    if gGlobalSyncTable.levelIndex == 0 then
        fullGameIndex = 1
        gGlobalSyncTable.level = fullGameSeq[fullGameIndex]
    else
        gGlobalSyncTable.level = gGlobalSyncTable.levelIndex
    end
end

function next_level()
    gGlobalSyncTable.transform = false
    fullGameIndex = fullGameIndex + 1
    gGlobalSyncTable.level = fullGameSeq[fullGameIndex]
end

function end_round()
    gGlobalSyncTable.level = 0
end

function record_time()

    local clearMessage = "."
    if gGlobalSyncTable.levelIndex == 0 then
        clearMessage = "All Courses completed in " .. convert_time(gGlobalSyncTable.timer, "speedrun")
    else
        clearMessage = get_level_name(levels[gGlobalSyncTable.levelIndex][2], 1, 1) .. " cleared in " .. convert_time(gGlobalSyncTable.timer, "speedrun")
    end

    packet_receive({packet="message", message=clearMessage})
    network_send(true, {packet="message", message=clearMessage})

    if soloRun then
        if
            mod_storage_load(game .. gGlobalSyncTable.levelIndex) == nil or
            gGlobalSyncTable.timer < tonumber(mod_storage_load(game .. gGlobalSyncTable.levelIndex))
        then
            mod_storage_save(game .. gGlobalSyncTable.levelIndex, tostring(gGlobalSyncTable.timer))
        end
    end

end

function all_stars_collected()
    local allStars = true
    for i = 0, levels[gGlobalSyncTable.level][3]-1 do
        if save_file_get_star_flags(get_current_save_file_num() - 1, levels[gGlobalSyncTable.level][2]-1) & (1 << i) == 0 then
            allStars = false
        end
    end
    return allStars
end

hook_event(HOOK_UPDATE, update)