------------------------
--BLUE COIN SWITCH FIX--
------------------------

local function switch(param, case_table)
    local case = case_table[param]
    if case then return case() end
    return nil
end

---Needed to be recreated
---@param parent Object
---@param model ModelExtendedId
---@param behaviorId BehaviorId
local function spawn_object(parent, model, behaviorId)
    local obj = spawn_non_sync_object(behaviorId, model, 0, 0, 0, nil)
    if obj == nil then return nil end

    obj_copy_pos_and_angle(obj, parent)
    return obj
end

---@type Vec3f
local blue_coin_switch_pos = {x = 0, y = 0, z = 0}
local blue_coin_switch_angle = 0

---@param obj Object
function custom_blue_coin_switch_init(obj)
    vec3f_set(blue_coin_switch_pos, obj.oPosX, obj.oPosY, obj.oPosZ)
    blue_coin_switch_angle = obj.oFaceAngleYaw
end

---@param obj Object
function custom_blue_coin_switch_loop(obj)
    if obj.oAction == BLUE_COIN_SWITCH_ACT_TICKING and obj.oTimer > 240 then
        spawn_non_sync_object(
            id_bhvBlueCoinSwitch,
            E_MODEL_BLUE_COIN_SWITCH,
            blue_coin_switch_pos.x, blue_coin_switch_pos.y, blue_coin_switch_pos.z,
            -- Redefinition necessary
            ---@param obj Object
            function (obj)
                obj.oFaceAngleYaw = blue_coin_switch_angle
            end
        )
    end
end

hook_behavior(id_bhvBlueCoinSwitch, OBJ_LIST_SURFACE, false, custom_blue_coin_switch_init, custom_blue_coin_switch_loop)

---@param obj Object
function custom_hidden_blue_coin_init(obj)
    obj_set_model_extended(obj, E_MODEL_BLUE_COIN)
    load_object_collision_model() -- Unneeded?

    obj.oFlags = (OBJ_FLAG_ACTIVE_FROM_AFAR | OBJ_FLAG_COMPUTE_DIST_TO_MARIO | OBJ_FLAG_UPDATE_GFX_POS_AND_ANGLE)
    obj_set_billboard(obj)

    obj.oInteractType = INTERACT_COIN
    obj.hitboxRadius = 100
    obj.hitboxHeight = 64
    obj.oDamageOrCoinValue = 5
    obj.oIntangibleTimer = 0
    obj.oAnimState = -1
end

---@param obj Object
function custom_hidden_blue_coin_loop(obj)
    obj.oAnimState = obj.oAnimState + 1
    switch(obj.oAction,
    {
        [HIDDEN_BLUE_COIN_ACT_INACTIVE] = function()
            -- Become invisible and intangible
            cur_obj_disable_rendering()
            cur_obj_become_intangible()

            -- Set action to HIDDEN_BLUE_COIN_ACT_WAITING after the blue coin switch is found.
            obj.oHiddenBlueCoinSwitch = cur_obj_nearest_object_with_behavior(get_behavior_from_id(id_bhvBlueCoinSwitch))

            if obj.oHiddenBlueCoinSwitch ~= nil then
                obj.oAction = obj.oAction + 1 -- Set to HIDDEN_BLUE_COIN_ACT_WAITING
            end
        end,
        [HIDDEN_BLUE_COIN_ACT_WAITING] = function()
            -- Wait until the blue coin switch starts ticking to activate.
            local blueCoinSwitch = obj.oHiddenBlueCoinSwitch

            if blueCoinSwitch.oAction == BLUE_COIN_SWITCH_ACT_TICKING then
                obj.oAction = obj.oAction + 1 -- Set to HIDDEN_BLUE_COIN_ACT_ACTIVE
            end
        end,
        [HIDDEN_BLUE_COIN_ACT_ACTIVE] = function()
            -- Become tangible
            cur_obj_enable_rendering()
            cur_obj_become_tangible()

            -- Delete the coin once collected
            if obj.oInteractStatus & INT_STATUS_INTERACTED ~= 0 then
                spawn_object(obj, E_MODEL_SPARKLES, id_bhvGoldenCoinSparkles)
                obj_mark_for_deletion(obj)
            end

            -- 200 frames of waiting and 20 2-frame blinks (for 240 frames total)
            if cur_obj_wait_then_blink(200, 20) ~= 0 then
                obj.header.gfx.node.flags = obj.header.gfx.node.flags &~ GRAPH_RENDER_INVISIBLE -- cur_obj_wait_then_blink causes this flag to be set so I take it off
                obj.oAction = obj.oAction - 2 -- Set back to HIDDEN_BLUE_COIN_ACT_INACTIVE
            end
        end
    })
end

hook_behavior(id_bhvHiddenBlueCoin, OBJ_LIST_LEVEL, true, custom_hidden_blue_coin_init, custom_hidden_blue_coin_loop)